"""Class PitchContext, which organizes computation of pitch context vectors by combining a Song and a class derived from ComputePitchContext.
This is the only class publicly available to the user."""

from __future__ import annotations

from fractions import Fraction

import numpy as np

from .song import Song
from .PCParameters import PCParameters
from .ComputePitchContext import ComputePitchContextBeats, ComputePitchContextNotes, ComputePitchContextScoretime
from .base40 import base40

#weighted pitch contect
class PitchContext:
    """Class for computing a weighted pitch context vector and keeping track of the parameters.
    
    Parameters
    ----------
    song : Song
        instance of the class Song, representing the song as MTCFeatures, with some
        additional features. By instanciating an object of this class, the following
        parameters have to be provided. Parameters with a default value can be ommitted.
    removeRepeats : boolean, default=True
        If True, skip all notes with repeated pitches.
    syncopes : boolean, default=False
        If True, take the highest metric weight DURING the span of the note.
        If False, take the metric weight at the onset time of the note.
    metric_weights : one of 'beatstrength', 'ima', 'imaspect', default='beatstrength'
        `beatstrength` : use beatstrength as computed by music21
        `ima` : use inner metric analysis weights (not yet implemented)
        `imaspect` : use inner metric analysis spectral weights (not yet implemented)
    accumulateWeight : boolean, default=False
        If true, represent the metric weight of a note by the sum of all metric weights
        in the beatstrength grid in the span of the note.
    context_type : one of 'scoretime', 'beats', 'notes', default='beats'
        scoretime: len_context is a float in units of quarterLength (length of one quarter note)
        beats: len_context is a float in units of beat (length of beat as computed by music21)
        notes: len_context is an integer, number of notes
    len_context: float or (float, float), int or (int, int), 'auto' or (...,'auto') or ('auto',...)
        length of the context. Value depends on context_type. See context_type.
        If a tuple is given, first in the tuple refers to preceding context and
        second in the tuple to the following context.
        'auto': the length of the context is determined automatically.
    len_context_params: dict, or (dict, dict) default = {}
        Named parameters for determination of context length. Depends on algorithm.
    use_metric_weights : boolean or (boolean, boolean), default=True
        Whether to weight the pitches in the conext by their metric weight.
        If a tuple is given, first in the tuple refers to preceding context and
        second in the tuple to the following context.
    use_distance_weights : boolean or (boolean, boolean), default=True
        If True, weight pithces in the context by their distance to the focus note.
        The weight is a linear function of the score distance to the focus note.
        The weight at the onset of the focus note is 1.0.
        The weight at the end of the context is set by `min_distance_weight`.
        If a tuple is given, first in the tuple refers to preceding context and
        second in the tuple to the following context.
    min_distance_weight : float of (float, float), default=0.0
        Distance weight at the border of the context.
        If a tuple is given, first in the tuple refers to preceding context and
        second in the tuple to the following context.
    include_focus : boolean or (boolean, boolean), default=True
        Whether to include the focus note in the context.
        If a tuple is given, first in the tuple refers to preceding context and
        second in the tuple to the following context.
    partial_notes : boolean, default=True
        If True, extend the PRE conext to the START of the first note within the context.
        This has consequences if the pre context starts IN a note.

    Attributes
    ----------
    params : dict
        Parameters for computing the WPC.
    ixs : list
        List of the indices of the notes that can be part of a context. This list is the single
        point of conversion to actual note indices within the melody.
    weightedpitch : numpy array
        Dimension is (length of ixs, 40). The first dimension corresponds to the note indices in
        `ixs`. The second dimension contains the metric weight of the corresponding note for the
        appropriate pitch in base 40 encoding.
    pitchcontext : numpy array
        Dimension is (length of ixs, 80). The first dimension corresponds to the note indices in
        `ixs`. The second dimension correpsonds to 40 pitches in the preceding context [:40] and
        40 pitches in the following context [40:]. Pitches are in base40 encoding.
    contexts_pre : list of lists
        Length is length of isx. For each note in `ixs`, `context_pre` contains a list of the
        indices of pitches in ixs that are part of the preceding context of the note.
    contexts_post : list of lists
        Length is length of isx. For each note in `ixs`, `context_post` contains a list of the
        indices of pitches in ixs that are part of the following context of the note.
    """

    def __init__(self, song, **inparams):
        self.song = song
        #contains params for actual contents of weightedpitch vector and weightedpitch context vector
        self.params = PCParameters(**inparams)
        #store for quick use
        self.epsilon = self.params.epsilon
        #do the computations
        #First compute a weight for each note
        self.weightedpitch, self.ixs = self.computeWeightedPitch()
        #create the proper computePitchContext object
        self.cpc = self.createComputePitchContext()
        #compute the pitch context
        self.pitchcontext, self.contexts_pre, self.contexts_post = self.computePitchContext()

    def createComputePitchContext(self):
        if self.params.context_type == 'beats':
            return ComputePitchContextBeats(self)
        elif self.params.context_type == 'notes':
            return ComputePitchContextNotes(self)
        elif self.params.context_type == 'scoretime':
            return ComputePitchContextScoretime(self)
        else:
            print("Error: Unsupported context type: " + self.params.context_type)

    def computeWeightedPitch(self):
        """Computes for every note a pitchvector (base40) with the (metric) weight of the note in the corresponding pitch bin.

        Returns
        -------
        numpy array, numpy array
            1. weightedpitch: Dimension is (length of ixs, 40). The first dimension corresponds to the note indices in
            `ixs`. The second dimension contains the metric weight of the corresponding note for the
            appropriate pitch in base 40 encoding.
            2 ixs: indices of the notes in the original melody.
        """
        #put param values in local variables for readibility
        removeRepeats = self.params.remove_repeats
        syncopes = self.params.syncopes
        metric_weights = self.params.metric_weights
        accumulateWeight = self.params.accumulate_weight

        if metric_weights in ['ima', 'imaspect']:
            raise Exception(f'{metric_weights} not yet implemented.')

        #local variables as proxy
        songinstance = self.song
        song = self.song.mtcsong

        onsettick = song['features']['onsettick']
        pitch40 = song['features']['pitch40']
        beatstrengthgrid = np.array(song['features']['beatstrengthgrid'])
        beatstrength = song['features']['beatstrength']
        song_length = songinstance.getSongLength()

        #find out which notes to keep (remove repeated notes)
        ixs = []
        if removeRepeats:
            p_prev=-1
            for ix, p40 in enumerate(song['features']['pitch40']):
                if p40 != p_prev:
                    ixs.append(ix)
                p_prev = p40
        else:
            ixs = list(range(song_length))

        #list to store the result
        weights = [0.0]*len(ixs)

        #on and offsets for selected notes:
        start_onsets = [onsettick[ix] for ix in ixs]
        stop_onsets = [song['features']['offsettick'][ix] for ix in ixs]

        if accumulateWeight:
            if syncopes:
                syncopes=False
                print("Warning: setting accumulateWeight implies syncopes=False.")
            #for each note make span of onsets:
            for ix, span in enumerate(zip(start_onsets, stop_onsets)):
                weights[ix] = sum(beatstrengthgrid[span[0]:span[1]])
        else:
            weights = [beatstrength[ix] for ix in ixs]
        
        if syncopes:
            for ix, span in enumerate(zip(start_onsets, stop_onsets)):
                maxbeatstrength = np.max(beatstrengthgrid[span[0]:span[1]])
                weights[ix] = maxbeatstrength

        song['features']['weights'] = [0.0] * song_length
        for ix, songix in enumerate(ixs):
            song['features']['weights'][songix] = weights[ix]

        weightedpitch = np.zeros( (len(ixs), 40) )
        for ix, songix in enumerate(ixs):
            p = pitch40[songix]
            w = weights[ix]
            weightedpitch[ix, (p-1)%40] = w
        return weightedpitch, ixs

    def getBeatinsongFloat(self):
        """Convert `beatinsong` from Fraction to float

        Returns
        -------
        numpy vector
            Length is length of `ixs`. numpy vector with beatinsong as float
        """
        song = self.song.mtcsong
        beatinsong_float = np.zeros( len(self.ixs) )
        for ix, song_ix in enumerate(self.ixs):
            beatinsong_float[ix] = float(Fraction(song['features']['beatinsong'][song_ix]))
        return beatinsong_float

    def computePitchContext(self):   
        """Compute for each note a pitchcontext vector

        Returns
        -------
        numpy arrays, list, list
            1. pichcontext: Dimension is (length of `ixs`, 80). The first dimension corresponds to the note indices in
            `ixs`. The second dimension correpsonds to 40 pitches in the preceding context [:40] and
            40 pitches in the following context [40:]. Pitches are in base40 encoding.
            2. contexts_pre: for each note the list of indices (ixs) of the notes in the preceding context.
            3. contexts_post: for each note the list of indices (ixs) of the notes in the following context.
        """
        #put param values in local variables for readibility
        use_metric_weights_pre = self.params.use_metric_weights_pre
        use_metric_weights_post = self.params.use_metric_weights_post
        use_distance_weights_pre = self.params.use_distance_weights_pre
        use_distance_weights_post = self.params.use_distance_weights_post
        
        #array to store the result
        pitchcontext = np.zeros( (len(self.ixs), 40 * 2) )

        #Lists for the indices of the contexts for each note
        contexts_pre = []
        contexts_post = []
        
        for ix in range(len(self.ixs)):
            #get context for the note (list of note indices)
            context_pre_ixs = self.cpc.computePreContext(ix, **self.params.len_context_params_pre)
            context_post_ixs = self.cpc.computePostContext(ix, **self.params.len_context_params_post)

            # print('context_pre', context_pre_ixs)
            # print('context_post', context_post_ixs)

            contexts_pre.append(context_pre_ixs)
            contexts_post.append(context_post_ixs)

            #compute distance-weights
            distance_weights_pre = np.ones(context_pre_ixs.shape)
            if use_distance_weights_pre:
                distance_weights_pre = self.cpc.computeDistanceWeightsPre(context_pre_ixs, ix)

            distance_weights_post = np.ones(context_post_ixs.shape)
            if use_distance_weights_post:
                distance_weights_post = self.cpc.computeDistanceWeightsPost(context_post_ixs, ix)

            #set metric weights to 1 if not use_metric_weights
            metric_weights_pre = self.weightedpitch[context_pre_ixs]
            if not use_metric_weights_pre:
                metric_weights_pre[metric_weights_pre>0] = 1.0

            metric_weights_post = self.weightedpitch[context_post_ixs]
            if not use_metric_weights_post:
                metric_weights_post[metric_weights_post>0] = 1.0

            # print("ix", ix, self.ixs[ix])
            # print("context_pre_ixs", context_pre_ixs)
            # print("distance_weights_pre", distance_weights_pre)
            # print("metric_weights_pre", metric_weights_pre)
            # print("self.weightedpitch[context_pre_ixs]", self.weightedpitch[context_pre_ixs])
            # print("length_context_post", length_context_post)
            # print("distance_weights_pre", distance_weights_pre)
            # print("distance_weights_post", distance_weights_post)
            #combine context into one vector

            pitchcontext_pre  = np.dot(distance_weights_pre, metric_weights_pre)
            pitchcontext_post = np.dot(distance_weights_post, metric_weights_post)
            
            #store result
            pitchcontext[ix,:40] = pitchcontext_pre
            pitchcontext[ix,40:] = pitchcontext_post

        return pitchcontext, contexts_pre, contexts_post

    def printReport(
        self,
        note_ix=None, #report single note. IX in original song, not in ixs
        **features, #any other values to report. key: name, value: array size len(ixs)
    ):
        """Returns a textual report with for each note the values of several features.

        For each note print
        - pitch and (metric) weight as computed by `self.computeWeightedPitch`
        - indices (in `self.ixs`) of notes in the preceding context
        - indices (in the MTC features) of notes in the preceding context
        - indices (in `self.ixs`) of notes in the following context
        - indices (in the MTC features) of notes in the following context
        - pitches and corresponding weights in the precedings context
        - pitches and corresponding wieghts in the following context
        - any other feature provided as keyword argument (see below)

        Parameters
        ----------
        note_ix : int, default None
            Only print the values the note at index `note_ix` in the original melody (not in `self.ixs`).
        **features  : keyword arguments
            any other feature to report. The keyword is the name of the feature, the value is a 1D array
            with the same length as `self.ixs`.

        Returns
        ----------
        str
            String containing the report.
        """
        output = []
        for ix in range(len(self.ixs)):
            if note_ix:
                if note_ix != self.ixs[ix]: continue
            pre_pitches = []
            post_pitches = []
            for p in range(40):
                if self.pitchcontext[ix,p] > 0.0:
                    pre_pitches.append((base40[p],self.pitchcontext[ix,p]))
            for p in range(40):
                if self.pitchcontext[ix,p+40] > 0.0:
                    post_pitches.append((base40[p], self.pitchcontext[ix,p+40]))
            pre_pitches = [str(p) for p in sorted(pre_pitches, key=lambda x: x[1], reverse=True)]
            post_pitches = [str(p) for p in sorted(post_pitches, key=lambda x: x[1], reverse=True)]
            output.append(f"note {self.ixs[ix]}, ix: {ix}")
            output.append(f"  pitch, weight: {self.song.mtcsong['features']['pitch'][self.ixs[ix]]}, {self.song.mtcsong['features']['weights'][self.ixs[ix]]}")
            output.append(f"  context_pre (ixs): {self.contexts_pre[ix]}")
            output.append(f"  context_pre (notes): {np.array(self.ixs)[self.contexts_pre[ix]]}")
            output.append(f"  context_post (ixs): {self.contexts_post[ix]}")
            output.append(f"  context_post (notes): {np.array(self.ixs)[self.contexts_post[ix]]}")
            output.append( "  pre:" + "\n       ".join(pre_pitches))
            output.append( "  post:"+ "\n        ".join(post_pitches))
            for name in features.keys():
                output.append(f"  {name}: {features[name][ix]}")
            output.append("")
        return '\n'.join(output)




"""Classes to compute the preceding and following contexts for a focus note.

A concrete model inherits from the abstract class ComputePitchContext.
"""

from abc import ABC, abstractmethod
from fractions import Fraction
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from .pitchcontext import PitchContext

class ComputePitchContext(ABC):
    def __init__(self, wpc: 'PitchContext'):
        super().__init__()
        self.wpc = wpc
        self.song = wpc.song
        self.params = wpc.params        

    def computePreContext(self, focus_ix, **context_params_pre):
        if self.params.len_context_pre_auto:
            return self.computePreContextAuto(focus_ix, **context_params_pre)
        else:
            return self.computePreContextFixed(focus_ix, **context_params_pre)

    def computePostContext(self, focus_ix, **context_params_post):
        if self.params.len_context_post_auto:
            return self.computePostContextAuto(focus_ix, **context_params_post)
        else:
            return self.computePostContextFixed(focus_ix, **context_params_post)

    @abstractmethod
    def computePreContextFixed(self, focus_ix, **context_params_pre):
        pass

    @abstractmethod
    def computePostContextFixed(self, focus_ix, **context_params_post):
        pass

    @abstractmethod
    def computePreContextAuto(self, focus_ix, **context_params_pre):
        pass

    @abstractmethod
    def computePostContextAuto(self, focus_ix, **context_params_post):
        pass
    
    @abstractmethod
    def computeDistanceWeightsPre(self, context_pre_ixs, focus_ix):
        pass

    @abstractmethod
    def computeDistanceWeightsPost(self, context_post_ixs, focus_ix):
        pass

    @abstractmethod
    def computeContextLengthPre(self, context_ixs):
        pass

    @abstractmethod
    def computeContextLengthPost(self, context_ixs):
        pass

class ComputePitchContextExpWindow(ComputePitchContext):
    def __init__(self, wpc: 'PitchContext'):
        super().__init__(wpc)

    def computePreContextFixed(self, focus_ix, **context_params_pre):
        context_pre_ixs = []
        return context_pre_ixs

    def computePostContextFixed(self, focus_ix, **context_params_post):
        context_post_ixs = []
        return context_post_ixs

    def computePreContextAuto(self, focus_ix, **context_params_pre):
        context_pre_ixs = []
        return context_pre_ixs

    def computePostContextAuto(self, focus_ix, **context_params_post):
        context_post_ixs = []
        return context_post_ixs

    def computeDistanceWeightsPre(self, context_pre_ixs, focus_ix):
        distance_weights_pre = []
        return distance_weights_pre

    def computeDistanceWeightsPost(self, context_post_ixs, focus_ix):
        distance_weights_post = []
        return distance_weights_post

    def computeContextLength(self, context_ixs):
        len_context = 0
        return len_context

    def computeContextLengthPre(self, context_ixs):
        pass

    def computeContextLengthPost(self, context_ixs):
        pass


class ComputePitchContextScoretime(ComputePitchContext):
    pass


class ComputePitchContextNotes(ComputePitchContext):
    pass


class ComputePitchContextBeats(ComputePitchContext):
    def __init__(self, wpc: 'PitchContext'):
        super().__init__(wpc)
        #compute some extra features. LENGTH: wpc.ixs
        self.songlength_beat = self.song.mtcsong['features']['beatinsong_float'][wpc.ixs[-1]] + float(Fraction(self.song.mtcsong['features']['beatfraction'][wpc.ixs[-1]])) - self.song.mtcsong['features']['beatinsong_float'][0]
        self.beatinsong = np.array([self.song.mtcsong['features']['beatinsong_float'][ix] for ix in wpc.ixs])
        self.beatinsong_next = np.append(self.beatinsong[1:],self.songlength_beat+self.beatinsong[0]) #first beatinsong might be negative (upbeat)

    def computePreContextFixed(self, focus_ix, **context_params_pre):
        beatoffset = self.beatinsong - self.beatinsong[focus_ix]
        len_context_pre = self.params.len_context_pre
        epsilon = self.params.epsilon
        #N.B. for some reason, np.where returns a tuple e.g: (array([], dtype=int64),)
        if self.params.include_focus_pre:
            context_pre_ixs = np.where(np.logical_and(beatoffset>=-(len_context_pre + epsilon), beatoffset<=0))[0]
        else:
            context_pre_ixs = np.where(np.logical_and(beatoffset>=-(len_context_pre + epsilon), beatoffset<0))[0]
        
        if self.params.partial_notes:
            if focus_ix>0: #skip first, has no context_pre
                #check wether context start at beginning of a note. If not, add previous note
                #print(context_pre[0][0],beatoffset[context_pre[0][0]],len_context)
                if context_pre_ixs.shape[0]>0:
                    if np.abs( beatoffset[context_pre_ixs[0]] + len_context_pre ) > epsilon:
                        if context_pre_ixs[0]-1 >= 0:
                            context_pre_ixs = np.insert(context_pre_ixs, 0, context_pre_ixs[0]-1)
                else:
                    if len_context_pre>epsilon:
                        context_pre_ixs = np.insert(context_pre_ixs, 0, focus_ix-1) #if context was empty, add previous note anyway        
        
        return context_pre_ixs

    def computePostContextFixed(self, focus_ix, **context_params_post):
        beatoffset = self.beatinsong - self.beatinsong[focus_ix]
        slicelength = self.beatinsong_next[focus_ix] - self.beatinsong[focus_ix]
        beatoffset_next = beatoffset - slicelength #set onset of next note to 0.0
        len_context_post = self.params.len_context_post
        epsilon = self.params.epsilon
        #N.B. for some reason, np.where returns a tuple e.g: (array([], dtype=int64),)
        if self.params.include_focus_post:
            context_post_ixs = np.where(np.logical_and(beatoffset>=0, beatoffset_next<(len_context_post - epsilon)))[0]   
        else: # ['both', 'post']
            #start context at END of note
            #do not include the note that starts AT the end of the context
            context_post_ixs = np.where(np.logical_and(beatoffset_next>=0, beatoffset_next<(len_context_post - epsilon)))[0]
        return context_post_ixs

    def computePreContextAuto(self, focus_ix, **context_params_pre):
        #handle params
        if 'threshold' in context_params_pre:
            threshold = context_params_pre['threshold']
        else:
            threshold = 1.0 #default
        if 'not_heigher_than_focus' in context_params_pre:
            not_heigher_than_focus = context_params_pre['not_heigher_than_focus']
        else:
            not_heigher_than_focus = True #default

        context_pre_ixs = []
        if self.params.include_focus_pre:
            context_pre_ixs.append(focus_ix)
        ixadd = focus_ix - 1
        while True:
            if ixadd < 0:
                break
            context_pre_ixs.append(ixadd)
            threshold_condition = np.sum(self.wpc.weightedpitch[ixadd]) >= threshold-self.params.epsilon
            if not_heigher_than_focus:
                rel_condition = np.sum(self.wpc.weightedpitch[ixadd]) >= np.sum(self.wpc.weightedpitch[focus_ix])
            else:
                rel_condition = False
            if threshold_condition or rel_condition:
                break
            ixadd = ixadd - 1
        context_pre_ixs.reverse()
        context_pre_ixs = np.array(context_pre_ixs, dtype=int)
        return context_pre_ixs

    def computePostContextAuto(self, focus_ix, **context_params_post):
        #handle params
        if 'threshold' in context_params_post:
            threshold = context_params_post['threshold']
        else:
            threshold = 1.0 #default
        if 'not_heigher_than_focus' in context_params_post:
            not_heigher_than_focus = context_params_post['not_heigher_than_focus']
        else:
            not_heigher_than_focus = True #default
        
        context_post_ixs = []
        if self.params.include_focus_post:
            context_post_ixs.append(focus_ix)
        ixadd = focus_ix + 1
        while True:
            if ixadd >= len(self.wpc.ixs):
                break
            context_post_ixs.append(ixadd)
            threshold_condition = np.sum(self.wpc.weightedpitch[ixadd]) >= threshold-self.params.epsilon
            if not_heigher_than_focus:
                rel_condition = np.sum(self.wpc.weightedpitch[ixadd]) >= np.sum(self.wpc.weightedpitch[focus_ix])
            else:
                rel_condition = False
            if threshold_condition or rel_condition: # stop criterion
                #if threshold, remove last note
                if threshold_condition:
                    context_post_ixs.pop()
                #stop
                break
            ixadd = ixadd + 1
        context_post_ixs = np.array(context_post_ixs, dtype=int)
        return context_post_ixs

    def computeDistanceWeightsPre(self, context_pre_ixs, focus_ix):
        beatoffset_previous = self.beatinsong - self.beatinsong[focus_ix]
        mindist = self.params.min_distance_weight_pre
        len_context_pre = self.computeContextLengthPre(context_pre_ixs)
        distance_weights_pre  = beatoffset_previous[context_pre_ixs] * (1.0-mindist)/len_context_pre + 1.0
        #set negative weights to zero:
        distance_weights_pre[distance_weights_pre<0.0] = 0.0
        # if focus_ix == 6:
        #     print("note 6")
        #     print(f"{beatoffset_previous=}")
        #     print(f"{mindist=}")
        #     print(f"{len_context_pre=}")
        #     print("(1.0-mindist)/len_context_pre = ", (1.0-mindist)/len_context_pre)
        #     print("pre distance weights for note 6: ", distance_weights_pre)
        return distance_weights_pre

    def computeDistanceWeightsPost(self, context_post_ixs, focus_ix):
        #TODO: make distinction wether focus note is part of context or not.
        beatoffset = self.beatinsong - self.beatinsong[focus_ix]
        slicelength = self.beatinsong_next[focus_ix] - self.beatinsong[focus_ix]
        beatoffset_next = beatoffset - slicelength #set onset of next note to 0.0
        mindist = self.params.min_distance_weight_post
        len_context_post = self.computeContextLengthPost(context_post_ixs)
        distance_weights_post = beatoffset_next[context_post_ixs] * -(1.0-mindist)/len_context_post + 1.0
        #set negative weights to zero:
        distance_weights_post[distance_weights_post<0.0] = 0.0
        #set max weight to one (if focus note in post context, weight of focus note > 1.0)
        distance_weights_post[distance_weights_post>1.0] = 1.0
        # if focus_ix == 6:
            # print("note 6")
            # print(f"{slicelength=}")
            # print(f"{beatoffset_next=}")
            # print(f"{mindist=}")
            # print(f"{len_context_post=}")
            # print("-(1.0-mindist)/len_context_post = ", -(1.0-mindist)/len_context_post)
            # print("post distance weights for note 6: ", distance_weights_post)
        return distance_weights_post

    def computeContextLength(self, context_ixs):
        if len(context_ixs) > 0:
            len_context = self.beatinsong_next[context_ixs[-1]] - self.beatinsong[context_ixs[0]]
        else:
            len_context = 0.0
        return len_context

    def computeContextLengthPre(self, context_ixs):
        if self.params.len_context_pre_auto:
            return self.computeContextLength(context_ixs)
        else:
            return self.params.len_context_pre

    def computeContextLengthPost(self, context_ixs):
        if self.params.len_context_post_auto:
            return self.computeContextLength(context_ixs)
        else:
            return self.params.len_context_post



"""Provides class Cong, which encapsulates details about the melodic data."""

import copy
from fractions import Fraction
from math import gcd
import subprocess
import tempfile
import os
import json

import numpy as np
import music21 as m21
m21.humdrum.spineParser.flavors['JRP'] = True #For triplets
from IPython import display

#from datetime import datetime
#print(__file__, datetime.now().strftime("%H:%M:%S"))

#Exception: parsing failed
class ParseError(Exception):
    def __init__(self, arg):
        self.args = arg
    def __str__(self):
        return repr(self.value)

def lcm(a, b):
    """Computes the lowest common multiple.
    
    Parameters
    ----------
    a : int
    b : int

    Returns
    ------
    int
        Lowest common multiple of a and b.
    """
    return a * b // gcd(a, b)

def fraction_gcd(x, y):
    """Computes the greatest common divisor as Fraction
    
    Parameters
    ----------
    x : Fraction
    y : Fraction

    Returns
    ------
    Fraction
        greatest common divisor of x and y as Fraction.
    """
    a = x.numerator
    b = x.denominator
    c = y.numerator
    d = y.denominator
    return Fraction(gcd(a, c), lcm(b, d))

class Song:
    """Class for containing data about a song. An object of the class `Song` holds
    feature values for each note as provided by MTCFeatures, additionally computed
    features for computing the weighted pitch vectors, a music21 object representing
    the score.
    
    Parameters
    ----------
    mtcsong : dict
        Dictionary with feature values of all notes of the song, as provided by
        MTCFeatures
    krnfilename : string
        Filename of a corresponding **kern file

    Attributes
    ----------
    reduced : bool
        True if this song object is derived from another song object by removing notes.
    mtcsong : dict
        Dictionary with feature values of all notes of the song, as provided by
        MTCFeatures. The values of 'onsettick' are replaced by newly computed onsets.
    krnfilename : string
        Filename of a corresponding **kern file
    s : music21 Stream
        music21 object representing the score of the song. The following operation have
        been performed: ties removal, padding of incomplete bars, grace note removal.
        The  indices of the notes in the resulting stream should corresponsd to the indices
        of the feature values in `self.mtcsong`
    onsets : list of ints
        list with onset times. onsets[n] is the onset of the (n+1)th note. The term onset
        refers to the time in the score at which the note starts (in music21 this is offset).
        The resolution is determined by the number of ticks per quarter note
        as computed by getResolution().
    beatstrength_grid : list
        Contains a beatstrength for each possible onset as computed by the music21 meter model.
        Onsets are indices in the list. beatstrength_grid[n] is the beatstrength for a note
        starting at n.
    """

    def __init__(self, mtcsong, krnfilename, s_in=None, beatstrength_grid_in=None):
        """Instantiate an object of class Song.

        Parameters
        ----------
        mtcsong : dict
            Dictionary of feature values and metadata of a song as provided by MTCFeatures
        krnfilename : str
            Full file name of a corresponding **kern file.
            If krnfilename==None: derive from existing Song object (s_in and beatstrength_grid_in need to be provided.)
        s_in : m21.Stream
            If provided, do not parse from .krn file, but use the provided stream instead.
        beatstrength_grid_in : list (float)
            If provided, do not compute from .krn file, but use provided grid instead.
        """
        self.mtcsong = copy.deepcopy(mtcsong)
        if krnfilename != None:
            self.reduced = False
            self.krnfilename = krnfilename
            self.s = self.parseMelody()
            self.onsets = self.getOnsets()
            self.beatstrength_grid = self.create_beatstrength_grid()
        else:
            self.reduced = True
            self.krnfilename = None
            self.s = copy.deepcopy(s_in)
            self.onsets = self.getOnsets()
            self.beatstrength_grid = copy.deepcopy(beatstrength_grid_in)
        self.songlength = None
        #always do this: recomputes features for reduced Song objects
        self.add_features()

    def getSongLength(self):
        """Returns the number of notes in the song

        Returns
        -------
        int
            Number of notes in the song
        """
        if self.songlength == None:
            self.computeSongLength()
        return self.songlength

    def computeSongLength(self):
        self.songlength = len(self.mtcsong['features']['pitch'])

    def getDurationUnit(self):
        """Returns a unit of note duration that is the greatest common divisor of all note durations.

        Returns
        -------
        Fraction
            Duration unit
        """
        sf = self.s.flat.notesAndRests
        unit = Fraction(sf[0].duration.quarterLength)
        for n in sf:
            unit = fraction_gcd(unit, Fraction(n.duration.quarterLength))
        return fraction_gcd(unit, Fraction(1,1)) # make sure 1 is dividable by the unit.denominator

    #return number of ticks per quarter note
    def getResolution(self) -> int:
        """Return the number of ticks per quarter note given the duration unit.

        Returns
        -------
        int
            number of ticks per quarter note.
        """
        unit = self.getDurationUnit()
        #number of ticks is 1 / unit (if that is an integer)
        ticksPerQuarter = unit.denominator / unit.numerator
        if ticksPerQuarter.is_integer():
            return int(unit.denominator / unit.numerator)
        else:
            print(self.s.filePath, ' non integer number of ticks per Quarter')
            return 0

    def getOnsets(self):
        """Returns a list of onsets (ints). Onsets are multiples of the duration unit.

        N.B. this replaces the values of the feature 'onsettick' in the MTC json with the
        newly computed values.

        Returns
        -------
        list of int
            Onset for each note.
        """
        ticksPerQuarter = self.getResolution()
        onsets = [int(n.offset * ticksPerQuarter) for n in self.s.flat.notes]
        #check whether same onsets in songfeatures
        assert len(self.mtcsong['features']['onsettick']) == len(onsets)
        #NB. If initial rests (e.g. NLB142326_01) all onsets are shifted wrt MTC
        # REPLACE MTC values of onsettics with onsets as computed here (anyway)
        for ix in range(len(onsets)):
            self.mtcsong['features']['onsettick'][ix] = onsets[ix]                
        return onsets

    # s : music21 stream
    def removeGrace(self, s):
        """Remove all grace notes from the music21 stream.

        Parameters
        ----------
        s : music21 Stream
            Object representing the score of the song.
        """
        #highest level:
        graceNotes = [n for n in s.recurse().notes if n.duration.isGrace]
        for grace in graceNotes:
            grace.activeSite.remove(grace)
        #if s is not flat, there will be Parts and Measures:
        for p in s.getElementsByClass(m21.stream.Part):
            #Also check for notes at Part level.
            #NLB192154_01 has grace note in Part instead of in a Measure. Might be more.
            graceNotes = [n for n in p.recurse().notes if n.duration.isGrace]
            for grace in graceNotes:
                grace.activeSite.remove(grace)
            for ms in p.getElementsByClass(m21.stream.Measure):
                graceNotes = [n for n in ms.recurse().notes if n.duration.isGrace]
                for grace in graceNotes:
                    grace.activeSite.remove(grace)
        
    # add left padding to partial measure after repeat bar
    def padSplittedBars(self, s):
        """Add padding to bars that originate from splitting a bar at a repeat sign. The second of the two
        resulting (partial) bars should have a padding equal to the lenght of the first (partial) bar in 
        order to obtain correct beatstrength values.

        Parameters
        ----------
        s : music21 Stream
            Object representing the score of the song.

        Returns
        -------
        music21 Stream
            s with padded bars
        """
        partIds = [part.id for part in s.parts] 
        for partId in partIds: 
            measures = list(s.parts[partId].getElementsByClass('Measure')) 
            for m in zip(measures,measures[1:]): 
                if m[0].quarterLength + m[0].paddingLeft + m[1].quarterLength == m[0].barDuration.quarterLength: 
                    m[1].paddingLeft = m[0].quarterLength 
        return s

    #N.B. contrary to the function currently in MTCFeatures (nov 2022), do not flatten the stream
    def parseMelody(self, stripTies=True):
        """Converts **kern to music21 Stream and do necessary preprocessing:
        - pad splitted bars
        - strip ties
        - remove grace notes
        The notes in the resulting score correspond 1-to-1 with the notes in MTCFeatures.

        Returns
        -------
        music21 Stream
            music21 score for the song.

        Raises
        ------
        ParseError
            Raised if the **kern file is unparsable.
        """
        try:
            s = m21.converter.parse(self.krnfilename)
        except m21.converter.ConverterException:
            raise ParseError(self.krnfilename)
        #add padding to partial measure caused by repeat bar in middle of measure
        s = self.padSplittedBars(s)
        if stripTies:
            s = s.stripTies()
        self.removeGrace(s)
        return s
    
    def getTieMap(self, s):
        """Create a mapping from indices in the mtc feature to indices in the score, skipping tied notes.
        """
        def untied(note):
            if note.tie == None:
                return True
            else:
                if note.tie.type == 'start':
                    return True
                else:
                    return False
        return [
            ix for ix, n in enumerate(s.flat.notes) if untied(n)
        ]
    
    #Add metric grid
    def create_beatstrength_grid(self):
        """Creates a vector with for each possible onset the beatstrength. The last onset corresponds to the
        end of the last note.

        Returns
        -------
        list (float)
            list with beatstrengths. The beatstrenght for onset n is beatstrength_grid[n].
        """
        beatstrength_grid = []
        unit = Fraction(1, self.getResolution()) #duration of a tick
        s_onsets = m21.converter.parse(self.krnfilename) #original score
        for p in s_onsets.getElementsByClass(m21.stream.Part):
            for m in p.getElementsByClass(m21.stream.Measure):
                offset = 0*unit
                while offset < m.quarterLength:
                    n = m21.note.Note("C").getGrace()
                    m.insert(offset, n)
                    beatstrength_grid.append(n.beatStrength)
                    offset += unit
        assert beatstrength_grid[self.onsets[-1]] == self.mtcsong['features']['beatstrength'][-1]
        return beatstrength_grid

    def add_features(self):
        """Adds a few features that are needed for computing pitch vectors. One value for each note.
        - make sure pitch40 exists. In new version of MTCFeatures this is absent. Take pitch40_Hewlett
        - syncope: True if the note is a syncope (there is a a higher metric weight in the span of the note than at the start of the note).
        - maxbeatstrength: the highest beatstrenght DURING the note.
        - offsets: offset tick of the note (first tick AFTER the note)
        - beatinsong_float: float representation of beatinsong
        - is final pitch: bool. True if the pitch does not change anymore (final pitch might be repeated)
        """
        #assert pitch40
        if not 'pitch40' in self.mtcsong['features'].keys():
            self.mtcsong['features']['pitch40'] = copy.deepcopy(self.mtcsong['features']['pitch40_hewlett'])
        self.mtcsong['features']['syncope'] = [False] * len(self.mtcsong['features']['pitch'])
        self.mtcsong['features']['maxbeatstrength'] = [0.0] * len(self.mtcsong['features']['pitch'])
        self.mtcsong['features']['beatstrengthgrid'] = self.beatstrength_grid
        beatstrength_grid_np = np.array(self.beatstrength_grid)
        #offsets
        self.mtcsong['features']['offsettick'] = [0]*self.getSongLength()
        ticksperquarter = self.getResolution()
        for ix in range(self.getSongLength()):
            duration = int( Fraction(self.mtcsong['features']['duration_frac'][ix]) * ticksperquarter )
            self.mtcsong['features']['offsettick'][ix] = self.mtcsong['features']['onsettick'][ix] + duration
        #maxbeatstrength
        for ix, span in enumerate(zip(self.onsets,self.mtcsong['features']['offsettick'])): # (onset, offset)
            self.mtcsong['features']['maxbeatstrength'][ix] = self.mtcsong['features']['beatstrength'][ix]
            if np.max(beatstrength_grid_np[span[0]:span[1]]) > self.mtcsong['features']['beatstrength'][ix]:
                self.mtcsong['features']['syncope'][ix] = True
                self.mtcsong['features']['maxbeatstrength'][ix] = np.max(beatstrength_grid_np[span[0]:span[1]])
        #final note:
        #self.mtcsong['features']['maxbeatstrength'][-1] = self.mtcsong['features']['beatstrength'][-1]
        #beatinsong_float
        self.mtcsong['features']['beatinsong_float'] = [float(Fraction(b)) for b in self.mtcsong['features']['beatinsong']]
        #isfinalpitch
        self.mtcsong['features']['isfinalpitch'] = [False] * self.getSongLength()
        finalpitch = self.mtcsong['features']['pitch40'][-1] % 40 #disregard octave
        for ix in reversed(range(self.getSongLength())):
            if self.mtcsong['features']['pitch40'][ix] % 40 == finalpitch:
                self.mtcsong['features']['isfinalpitch'][ix] = True
            else:
                break
        #startfinalpitch
        self.mtcsong['features']['startfinalpitch'] = [False] * self.getSongLength()
        for ix in reversed(range(self.getSongLength()-1)):
            if self.mtcsong['features']['pitch40'][ix] % 40 != finalpitch:
                self.mtcsong['features']['startfinalpitch'][ix+1] = True
                break
        #highestweight: highest weight seen in preceeding notes during the sequence of FINAL pitch only
        self.mtcsong['features']['highestweight'] = [0.0] * self.getSongLength()
        for ix in range(self.getSongLength()):
            if not self.mtcsong['features']['isfinalpitch'][ix]:
                continue
            if self.mtcsong['features']['startfinalpitch'][ix]:
                continue
            #now we are sure that we are 'in' the final tone
            if self.mtcsong['features']['beatstrength'][ix-1] > self.mtcsong['features']['highestweight'][ix-1]:
                self.mtcsong['features']['highestweight'][ix] = self.mtcsong['features']['beatstrength'][ix-1]
            else:
                self.mtcsong['features']['highestweight'][ix] = self.mtcsong['features']['highestweight'][ix-1]


    def getReducedSong(self, ixs_remove, prolong_previous=False):
        """Create a new Song object without the notes in ixs_removed.

        Parameters
        ----------
        ixs_remove : list (int)
            Contains the indices of the notes that need to be removed.
        prolong_previous : bool
            If True, replace the to-be-removed note with a prolongation of the previous note.
            (Because of a bug in music21, this currently does not work if the song contains tuplets.)
            If False, replace the to-be-removed note with a rest.
            (Because of a bug in music21, this currently does not work if the song contains tuplets.)
        
        Returns
        -------
        Song object
            Song object
        """
        #create deep copy of m21 stream and beatstrength_grid
        s_new = copy.deepcopy(self.s)
        mtcsong_new = copy.deepcopy(self.mtcsong)
        beatstrength_grid_new = copy.deepcopy(self.beatstrength_grid)

        ixs_remove = sorted(ixs_remove)

        #Replace notes with prolongation of previous note (with ties, and then stripTies())
        #BUT not when the previous symbol is a rest!
        s_new_list = list(s_new.flat.notes)
        for ix in ixs_remove:

            #for now, leave 0 in. Subsequent notes might have to be removed as well
            if ix == 0:
                continue

            #get note
            n = s_new_list[ix]

            if prolong_previous:
                #check whether previous symbol is a rest
                prev_symbol = n.previous() #seems to work. It takes the right stream.
                while not 'Rest' in prev_symbol.classSet and not 'Note' in prev_symbol.classSet:
                    prev_symbol = prev_symbol.previous()
                if prev_symbol == None:
                    print("Beginning of stream encountered. Should not happen.")
                else:
                    if prev_symbol.isRest:
                        site = n.sites.getSitesByClass('Measure')[0] #assume it is the first (and only)
                        site.remove(n)
                        continue

                #previous symbol is not a rest
                n_prev = s_new_list[ix-1] #exists
                p_new = copy.deepcopy(s_new_list[ix-1].pitch) #take the pitch object of previous note
                s_new_list[ix].pitch = p_new
                #add ties
                if n_prev.tie == None:
                    n_prev.tie = m21.tie.Tie('start')
                else: #previous has tie (must be 'stop')
                    if n_prev.tie.type == 'stop':
                        n_prev.tie = m21.tie.Tie('continue')
                    elif n_prev.tie.type == 'continue':
                        print('Tie of previous note should not be of type continue')
                    else:
                        print('Tie of previous note should not be of type ', n_prev.tie.type)
                n.tie = m21.tie.Tie('stop')
            else: # just remove the note. Will be filled with rests later
                site = n.sites.getSitesByClass('Measure')[0] #assume it is the first (and only)
                site.remove(n)

        #update offsets and durations based on ties:
        #backward, and stop at 1 (TODO)
        if prolong_previous:
            for ix in reversed(range(1, len(s_new_list))):
                n = s_new_list[ix]
                n_prev = s_new_list[ix-1]
                if n.tie != None:
                    if n.tie.type != 'start':
                        #copy offset of tied note to previous note
                        mtcsong_new['features']['offsettick'][ix-1] = mtcsong_new['features']['offsettick'][ix]
                        #add duration to previous note
                        mtcsong_new['features']['duration_frac'][ix-1] = str( Fraction(mtcsong_new['features']['duration_frac'][ix-1]) + Fraction(mtcsong_new['features']['duration_frac'][ix]) )

        #replace removed notes with rests

        for m in s_new.recurse(classFilter=('Measure')):
            m.makeRests(inPlace=True, fillGaps=True, timeRangeFromBarDuration=True)

        #remove ties
        if prolong_previous:
            s_new = s_new.stripTies()

        #now remove first note (if 0 in ixs_remove)
        #also fix in mtcsong_new (below)
        #TODO

        #go over all features in mtcsong_new and remove... this will make a mess
        #for offsettick
        for feat in mtcsong_new['features'].keys():
            for ix in reversed(ixs_remove):
                if ix != 0: #TODO See above
                    mtcsong_new['features'][feat].pop(ix)
        #list of onsets might need to be divided by the gcd
        onsetgcd = np.gcd.reduce(mtcsong_new['features']['onsettick'])
        if onsetgcd > 1:
            for ix in range(len(mtcsong_new['features']['onsettick'])):
                mtcsong_new['features']['onsettick'][ix] = int( mtcsong_new['features']['onsettick'][ix] / onsetgcd )

        song_new = Song(mtcsong_new, None, s_in=s_new, beatstrength_grid_in=beatstrength_grid_new)
        return song_new


    def getColoredSong(self, colordict, lyrics=None, lyrics_ixs=None, title=None):
        """Create a new music21 stream with notes colored according to `colordict`.

        Parameters
        ----------
        colordict : dict
            The keys are the colors, the values the indices of the notes with that color. E.g. {'red':[0,10,11],'grey':[-1]}
            colors notes at indices 0, 10, and 11 red, and the last note grey.
        lyrics : list (str) OR list of lists (str)
            Use the items in this list as lyrics, one string for each note.
            If a list of lists is provided, these will be printed as different lines of lyrics.
        lyrics_ixs : list (int)
            If given, put the lyrics at the given indices in the song. Should have the same length as lyrics.
            If not given, lyrics should have the same length as the song.

        Returns
        -------
        music21 Stream
            music21 Stream.
        """
        if self.reduced:
            s = copy.deepcopy(self.s)
        else:
            s = self.parseMelody(stripTies=False)
        if title != None:
            s.metadata.title = title
        tie_map = self.getTieMap(s)
        #check for right length #if so, assume notes correspond with features
        assert self.getSongLength() == len(tie_map)
        for color, ixs in colordict.items():
            for ix in ixs:
                ix = tie_map[ix]
                s.flat.notes[int(ix)].style.color = color
        #check whether continuation of ties needs to be colored
        for ix, n in enumerate(s.flat.notes):
            if n.tie != None:
                if n.tie.type != 'start':
                    s.flat.notes[int(ix)].style.color = s.flat.notes[int(ix-1)].style.color
        #add index of note as lyric
        if lyrics == None:
            lyrics = [str(ix) for ix in range(len(tie_map))]
            lyrics_ixs = list(range(len(tie_map)))
        #are more lines of lyrics provided?
        multipleLines = False
        if lyrics != None:
            if type(lyrics[0]) == list:
                multipleLines = True
        if lyrics_ixs == None:
            if not multipleLines and len(lyrics) == len(tie_map):
                lyrics_ixs = list(range(len(tie_map)))
            elif multipleLines and len(lyrics[0]) == len(tie_map):
                lyrics_ixs = list(range(len(tie_map)))
            else:
                print('Provide indices for the lyrics')
                for ix, n in enumerate(s.flat.notes):
                    n.lyric = None
                return s
        #make sure it is a list (to use .index)
        lyrics_ixs = list(lyrics_ixs)
        for ix, n in enumerate(s.flat.notes):
            try:
                ix = tie_map.index(ix)
            except ValueError:
                continue
            n.lyric = None
            try:
                lyricix = lyrics_ixs.index(ix)
            except ValueError:
                continue
            if multipleLines:
                for line in lyrics:
                    n.addLyric(line[lyricix])
            else:
                n.addLyric(lyrics[lyricix])
        return s
    
    #we need to repair lily generated by m21 concerning color
    #\override Stem.color -> \once\override Stem.color
    #\override NoteHead.color -> \once\override NoteHead.color

    def repairLyline(self, line):
        """Corrects possbile errors in a line of the Ly export:
        - Note coloring should be done once.
        - Melisma are somehow following beams (instead of slurs)
        - Beaming is wrong. 16th and 32th etc notes get 1 beam.
        
        Parameters
        ----------
        line : str
            a line of a generated lilypond file.

        Returns
        -------
        str
            corrected line
        """
        line = line.replace("\\override Stem.color","\\once\\override Stem.color")
        line = line.replace("\\override NoteHead.color","\\once\\override NoteHead.color")
        line = line.replace("\\include \"lilypond-book-preamble.ly\"","")

        line = line.replace("\\set stemLeftBeamCount = #1", "")        
        line = line.replace("\\set stemRightBeamCount = #1", "")        
        return line
    
    def formatAndRepairLy(self, filename):
        """Go over a lilypond file, and correct the lines (see `self.repairLyline`).
        Clear tagline.
        Set indent of first system to 0.

        Parameters
        ----------
        filename : str
            Filename of the lilypond file to repair.
        """
        with open(filename,'r') as f:
            lines = [self.repairLyline(l) for l in f.readlines()]
        lines = lines + [ f'\paper {{ tagline = "" \nindent=0}}']
        with open(filename,'w') as f:
            f.writelines(lines)

    def insertFilenameLy(self, filename):
        """Insert a header with the filename into a lilypond source file.

        Parameters
        ----------
        filename : str
            Filename of the lilypond file to process.
        """
        with open(filename,'r') as f:
            lines = [l for l in f.readlines()]
        with open(filename,'w') as f:
            for l in lines:
                if "\\score" in l:
                    f.write(f'\\header {{ opus = "{filename}" }}\n\n')
                f.write(l)

    def insertSystemSpacingLy(self, filename):
        """Insert a paper block with system-system spacing into lilypond source file.

        Parameters
        ----------
        filename : str
            Filename of the lilypond file to process.
        """
        with open(filename,'r') as f:
            lines = [l for l in f.readlines()]
        with open(filename,'w') as f:
            for l in lines:
                f.write(l)
                if "\\version" in l:
                    f.write( "\paper { system-system-spacing.basic-distance = #16 }\n" )

    def createColoredPDF(self, colordict, outputpath, filebasename=None, showfilename=True, lyrics=None, lyrics_ixs=None, title=None):
        """Create a pdf with a score with colored notes.

        Parameters
        ----------
        colordict : dict
            The keys are the colors, the values the indices of the notes with that color. E.g. {'red':[0,10,11],'grey':[-1]}
            colors notes at indices 0, 10, and 11 red, and the last note grey.
        outputpath : str
            name of the output directory
        filebasename : str, default None
            basename of the pdf file to generate (without .pdf). If None, the identifier of the song as provided by
            MTCFeatures is used as file name.
        showfilename : bool, default True
            Include the filename in the pdf (lilypond opus header).

        Returns
        -------
        path-like object
            Full path of the generated pdf.
        """
        if filebasename == None:
            filebasename = self.mtcsong['id']
        s = self.getColoredSong(colordict, lyrics=lyrics, lyrics_ixs=lyrics_ixs, title=title)
        s.write('lily', os.path.join(outputpath, filebasename+'.ly'))
        self.formatAndRepairLy(os.path.join(outputpath, filebasename+'.ly'))
        if showfilename:
            self.insertFilenameLy(os.path.join(outputpath, filebasename+'.ly'))
        self.insertSystemSpacingLy(os.path.join(outputpath, filebasename+'.ly'))
        output = subprocess.run(["lilypond", os.path.join(outputpath, filebasename+'.ly')], cwd=outputpath, capture_output=True)
        return os.path.join(outputpath, filebasename+'.pdf')

    def createColoredPNG(self, colordict, outputpath, filebasename=None, showfilename=True, lyrics=None, lyrics_ixs=None, title=None):
        """Create a png with a score with colored notes.

        Parameters
        ----------
        colordict : dict
            The keys are the colors, the values the indices of the notes with that color. E.g. {'red':[0,10,11],'grey':[-1]}
            colors notes at indices 0, 10, and 11 red, and the last note grey.
        outputpath : str
            name of the output directory
        filebasename : str, default None
            basename of the png file to generate (without .png). If None, the identifier of the song as provided by
            MTCFeatures is used as file name.
        showfilename : bool, default True
            Include the filename in the png (lilypond opus header).

        Returns
        -------
        path-like object
            Full path of the generated png.
        """
        pdf_fn = self.createColoredPDF(colordict, outputpath, filebasename, showfilename, lyrics=lyrics, lyrics_ixs=lyrics_ixs, title=title)
        png_fn = pdf_fn.replace('.pdf','.png')
        output = subprocess.run(['magick', '-density', '100', pdf_fn, '-alpha', 'Remove', '-trim', png_fn], cwd=outputpath, capture_output=True)
        return png_fn
    
    def showColoredPNG(self, colordict, outputpath, filebasename=None, showfilename=True, lyrics=None, lyrics_ixs=None, title=None):
        """Show a png with a score with colored notes. For use in a Jupyter notebook.

        Parameters
        ----------
        colordict : dict
            The keys are the colors, the values the indices of the notes with that color. E.g. {'red':[0,10,11],'grey':[-1]}
            colors notes at indices 0, 10, and 11 red, and the last note grey.
        outputpath : str
            name of the output directory
        filebasename : str, default None
            basename of the png file to generate (without .png). If None, the identifier of the song as provided by
            MTCFeatures is used as file name.
        showfilename : bool, default True
            Include the filename in the png (lilypond opus header).
        """
        png_fn = self.createColoredPNG(colordict, outputpath, filebasename, showfilename, lyrics=lyrics, lyrics_ixs=lyrics_ixs, title=title)
        display.display(display.Image(png_fn))

    def showPNG(self, lyrics=None, lyrics_ixs=None):
        """Show a png with a score of the song. For use in a Jupyter notebook.
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            self.showColoredPNG({}, tmpdirname, showfilename=False, lyrics=lyrics, lyrics_ixs=lyrics_ixs)
    
    def createPNG(self, outputpath, filebasename=None, showfilename=False, lyrics=None, lyrics_ixs=None, title=None):
        return self.createColoredPNG({}, outputpath, filebasename=filebasename, showfilename=showfilename, lyrics=lyrics, lyrics_ixs=lyrics_ixs, title=title)


    def writeMTCJSON(self, outputpath, filebasename=None):
        if filebasename == None:
            filebasename = self.mtcsong['id']        
        with open(os.path.join(outputpath,filebasename+'.json'), 'w') as f:
            json.dump(self.mtcsong, f)
    
    def writePerSymbolJSON(self, outputpath, filebasename=None):
        if filebasename == None:
            filebasename = self.mtcsong['id']
        #build per symbol representation
        symbols = []
        for ix in range(self.getSongLength()):
            symbol = {}
            #due to a bug in MTCFeatures, there is no octave information in the pitch40 feature
            #figure out the octave from the midi pitch, and correct the pitch40 value
            #take 1-based pitch40. I.e., the first octave is [1,40] (rather than [0,39])
            #midi middle C octave = 5 (sixth octave) = base40 octave 4 (fourth octave)
            octave = self.mtcsong['features']['midipitch'][ix] // 12
            pitch40 = 40 * (octave-2) + ((self.mtcsong['features']['pitch40'][ix] + 1) % 40) - 1  #still do mapping to base octave for pitch40 first, to make this future-proof.
            symbol["pitch40"] = pitch40
            symbol["onset"] = self.mtcsong['features']['onsettick'][ix]
            symbol["phrase"] = self.mtcsong['features']['phrase_ix'][ix]
            symbol["ima"] = self.mtcsong['features']['imaweight'][ix]
            symbol["phrasepos"] = self.mtcsong['features']['phrasepos'][ix]
            symbol["beatstrength"] = self.mtcsong['features']['beatstrength'][ix]
            symbols.append(symbol)
        songdictINNER = {}
        songdictINNER['symbols'] = symbols
        songdict = {}
        songdict[filebasename] = songdictINNER
        with open(os.path.join(outputpath,filebasename+'.json'), 'w') as f:
            json.dump(songdict, f)


"""Class to manage parameter configuration."""

from dataclasses import dataclass, field
import typing

@dataclass()
class PCParameters():
    #For computing note weights
    remove_repeats: bool = True
    syncopes: bool = True
    metric_weights: str = 'beatstrength'
    accumulate_weight: bool = False

    #For computing weighted pitch context vectors
    context_type: str = 'beats'

    len_context: 'typing.Any' = None
    len_context_pre: 'typing.Any' = 'auto'
    len_context_pre_auto: bool = True
    len_context_post: 'typing.Any' = 'auto'
    len_context_post_auto: bool = True

    len_context_params: 'typing.Any' = None
    len_context_params_pre: dict = field(default_factory=lambda : {})
    len_context_params_post: dict = field(default_factory=lambda : {})

    use_metric_weights: bool = None
    use_metric_weights_pre: bool = True
    use_metric_weights_post: bool = True

    use_distance_weights: bool = None
    use_distance_weights_pre: bool = True
    use_distance_weights_post: bool = True

    min_distance_weight: float = None
    min_distance_weight_pre: float = 0.0
    min_distance_weight_post: float = 0.0

    include_focus: bool = None
    include_focus_pre: bool = True
    include_focus_post: bool = True

    partial_notes: bool = True
    epsilon: float = 1.0e-4

    def __post_init__(self):
        #make sure these exist
        self.len_context_pre_auto = False
        self.len_context_post_auto = False

        #split into pre an post contexts       
        if self.len_context != None:
            self.len_context_pre = self.len_context
            self.len_context_post = self.len_context
            self.len_context = None
       
        if self.len_context_pre == 'auto':
            self.len_context_pre_auto = True
            self.len_context_pre = None
       
        if self.len_context_post == 'auto':
            self.len_context_post_auto = True
            self.len_context_post = None

        if self.len_context_params != None:
            self.len_context_params_pre = self.len_context_params
            self.len_context_params_post = self.len_context_params
            self.len_context_params = None

        if self.use_metric_weights != None:
            self.use_metric_weights_pre = self.use_metric_weights
            self.use_metric_weights_post = self.use_metric_weights
            self.use_metric_weights = None
       
        if self.use_distance_weights != None:
            self.use_distance_weights_pre = self.use_distance_weights
            self.use_distance_weights_post = self.use_distance_weights
            self.use_distance_weights = None
       
        if self.min_distance_weight != None:
            self.min_distance_weight_pre = self.min_distance_weight
            self.min_distance_weight_post = self.min_distance_weight
            self.min_distance_weight = None
       
        if self.include_focus != None:
            self.include_focus_pre = self.include_focus
            self.include_focus_post = self.include_focus
            self.include_focus = None



"""Models using the weighted pitch context vector."""

import numpy as np
from numpy.linalg import norm
from numpy import inf
from matplotlib import pyplot as plt
from copy import deepcopy

from .pitchcontext import PitchContext
from .song import Song
from .base40 import base40

#from datetime import datetime
#print(__file__, datetime.now().strftime("%H:%M:%S"))

#distances
def cosineSim(v1, v2):
    """Cosine Similarity of `v1` and `v2`, both 1D numpy arrays"""
    return np.dot(v1,v2)/(norm(v1)*norm(v2))

def normalizedCosineSim(v1, v2):
    """Cosine Similarity of `v1` and `v2`, both 1D numpy arrays. Scaled between 0 and 1."""
    return (1.0 + np.dot(v1,v2)/(norm(v1)*norm(v2))) / 2.0

def normalizedCosineDist(v1, v2):
    """One minus the normalized cosine similarity"""
    return 1.0-normalizedCosineSim(v1, v2)

def sseDist(v1, v2):
    """Sum of squared differences between `v1` and `v2`, both 1D numpy arrays."""
    return np.sum((v1-v2)**2)


#find out how dissonant a note is in its context
#consonant in base40:
# perfect prime : Dp = 0
# minor third : Dp = 11
# major third : Dp = 12
# perfect fourth : Dp = 17
# perfect fifth : Dp = 23
# minor sixth : Dp = 28
# major sixth : Dp = 29

def computeDissonance(
    song : Song, 
    wpc : PitchContext,
    combiner=np.maximum,
    normalizecontexts = False,
    consonants40 = [0, 11, 12, 17, 23, 28, 29]
):
    """
    Computes for each note the dissonance of the note given its context.

    Parameters
    ----------
    song : Song
        An instance of the Song class.
    wpd : WeightedPitchContext
        An instance of the WeightedPitchContext class, containing a weighted pitch context vector for each note.
    combiner : function of two 1D numpy arrays, default=numpy.maximum
        Combines the dissonance of preceding context and dissonance of following context in one value.
        Default: take the maximum.
    normalizecontexts : bool, default=False
        Normalize (sum-1.0) the context vectors before computing dissonance
    consonants : list of ints
        Intervals in base40 pitch encoding that are considered consonant.
    
    Returns
    -------
    dissonance_pre, dissonance_post, dissonance_context : numpy 1D arrays
        with a dissonance level for each note, respective the dissonance within the preceding context, the
        dissonance within the following context, and the dissonance within the full context.
    """
    song_length = len(wpc.ixs)

    dissonants = np.ones( (40,) )
    dissonants[consonants40] = 0.0

    #store result
    dissonance_pre = np.zeros( song_length )
    dissonance_post = np.zeros( song_length )
    dissonance_context = np.zeros( song_length )

    for ix, context in enumerate(wpc.pitchcontext): #go over the notes...

        pitch40 = song.mtcsong['features']['pitch40'][wpc.ixs[ix]]-1

        #make copy of context
        context = np.copy(context)

        #normalize contexts: sum of context is 1.0 (zero stays zero)
        if normalizecontexts:
            if np.sum(context[:40]) > 0.0:
                context[:40] = context[:40] / np.sum(context[:40])
            if np.sum(context[40:]) > 0.0:
                context[40:] = context[40:] / np.sum(context[40:])


        intervals_pre  = np.roll(context[:40], -pitch40)
        intervals_post = np.roll(context[40:], -pitch40)

        dissonance_pre[ix] = np.sum(np.multiply(intervals_pre, dissonants))
        dissonance_post[ix] = np.sum(np.multiply(intervals_post, dissonants))

        #if context is empty: value should be np.nan
        if len(wpc.contexts_pre[ix]) == 0:
            dissonance_pre[ix] = np.nan
        if len(wpc.contexts_post[ix]) == 0:
            dissonance_post[ix] = np.nan

    #combine pre and post context
    dissonance_context = combiner(dissonance_pre, dissonance_post)

    return dissonance_pre, dissonance_post, dissonance_context


def computeConsonance(
    song : Song, 
    wpc : PitchContext,
    combiner=np.minimum,
    normalizecontexts = False,
    consonants40 = [0, 11, 12, 17, 23, 28, 29]
):
    """
    Computes for each note the consonance of the note given its context.

    Parameters
    ----------
    song : Song
        An instance of the Song class.
    wpd : WeightedPitchContext
        An instance of the WeightedPitchContext class, containing a weighted pitch context vector for each note.
    combiner : function of two 1D numpy arrays, default=numpy.maximum
        Combines the consonance of preceding context and consonance of following context in one value.
        Default: take the maximum.
    normalizecontexts : bool, default=False
        Normalize (sum-1.0) the context vectors before computing consonance
    consonants : list of ints
        Intervals in base40 pitch encoding that are considered consonant.
    
    Returns
    -------
    consonance_pre, consonance_post, consonance_context : numpy 1D arrays
        with a consonance level for each note, respective the consonance within the preceding context, the
        consonance within the following context, and the consonance within the full context.
    """
    song_length = len(wpc.ixs)

    consonants = np.zeros( (40,) )
    consonants[consonants40] = 1.0

    #store result
    consonance_pre = np.zeros( song_length )
    consonance_post = np.zeros( song_length )
    consonance_context = np.zeros( song_length )

    for ix, context in enumerate(wpc.pitchcontext): #go over the notes...

        pitch40 = song.mtcsong['features']['pitch40'][wpc.ixs[ix]]-1

        #make copy of context
        context = np.copy(context)

        #normalize contexts: sum of context is 1.0 (zero stays zero)
        if normalizecontexts:
            if np.sum(context[:40]) > 0.0:
                context[:40] = context[:40] / np.sum(context[:40])
            if np.sum(context[40:]) > 0.0:
                context[40:] = context[40:] / np.sum(context[40:])

        intervals_pre  = np.roll(context[:40], -pitch40)
        intervals_post = np.roll(context[40:], -pitch40)

        consonance_pre[ix] = np.sum(np.multiply(intervals_pre, consonants))
        consonance_post[ix] = np.sum(np.multiply(intervals_post, consonants))

        #if context is empty: value should be np.nan
        if len(wpc.contexts_pre[ix]) == 0:
            consonance_pre[ix] = np.nan
        if len(wpc.contexts_post[ix]) == 0:
            consonance_post[ix] = np.nan

    #combine pre and post context
    consonance_context = combiner(consonance_pre, consonance_post)

    return consonance_pre, consonance_post, consonance_context


def computePrePostDistance(
    song : Song,
    wpc : PitchContext,
    vectorDist=normalizedCosineDist
):
    """Computes for each note the distance between the preceding and the following context.

    Parameters
    ----------
    song : Song
        Ojbect with song data.
    wpc : PitchContext
        Object with pitch context data
    vectorDist : function, default=normalizedCosineDist
        Function to compute the distance between two 1D vectors

    Returns
    -------
    numpy array
        1D numpy array with a distance value for each note.
    """
    res = np.zeros( len(wpc.pitchcontext) )
    for ix in range(len(wpc.pitchcontext)):
        res[ix] = vectorDist(wpc.pitchcontext[ix,:40], wpc.pitchcontext[ix,40:])
    return res

def computeNovelty(
    song: Song,
    wpc : PitchContext,
):
    """Computes for each note the 'novelty' of the following context with respect to the preceding context.
    Novelty for a pitch is computed as the percentual contribution of the following pitch value to the total
    of the preceding and following values.
    The overall novelty value is the average of novelty values of all pitches.

    Parameters
    ----------
    song : Song
        Ojbect with song data.
    wpc : PitchContext
        Object with pitch context data

    Returns
    -------
    numpy array
        1D numpy array with a novelty value for each note.
    """
    novelty = np.zeros( len(wpc.pitchcontext) )
    for ix in range(len(wpc.pitchcontext)):
        total = wpc.pitchcontext[ix,:40] + wpc.pitchcontext[ix,40:]
        new = wpc.pitchcontext[ix,40:]
        total[new==0] = 1
        perc  = new / total
        novelty[ix] = np.average(perc[perc>0])
    return novelty

def computeUnharmonicity(
    song: Song,
    wpc : PitchContext,
    dissonance: np.array,
    consonance: np.array,
    beatstrength_treshold: float,
    lastnoteharmonic: bool = False,
    epsilon: float = 10e-4
):
    """Computes for each note the degree to which it is 'unharmonic'.
    The unharmonicity value is the dissonance minus the consonance of a note in its context, with lower boundary of zero.
    Each note with beatstrength lower than beatstrength_threshold, and dissonant in its context is considered 'unharmonic' (value: 0.0).
    If parameter lastnoteharmonic is True, the last note gets unharmonicity value 0.0

    Parameters
    ----------
    song : Song
        Ojbect with song data.
    wpc : PitchContext
        Object with pitch context data
    dissonance : 1D numpy array
        Dissonance values as returned by computeDissonance
    consonance : 1D numpy array
        Consonance values as returned by computeConsonance
    beatstrength_threshold : float
        Consider all notes with beatstrength >= beatstrength_threshold as 'harmonic' (value 0.0).
    lastnoteharmonic : bool
        If True, always consider the last note of the melody as 'harmonic' (value 0.0).
        
    Returns
    -------
    numpy array
        1D numpy array with a unharmonicity value for each note.
    """
    beatstrength = song.mtcsong['features']['beatstrength']
    unharmonicity = np.zeros( len(wpc.pitchcontext) )
    for ix in range(len(wpc.pitchcontext)):
        if beatstrength[wpc.ixs[ix]] < beatstrength_treshold - epsilon:
            #unharmonicity[ix] = max ( dissonance[ix] - consonance[ix], 0.0 )
            if consonance[ix] == 0.0:
                unharmonicity[ix] = 10.0
            else:
                unharmonicity[ix] = max ( dissonance[ix] - consonance[ix], 0.0)
                #unharmonicity[ix] = dissonance[ix] / consonance[ix]
    #do last note separately
    if lastnoteharmonic:
        unharmonicity[-1] = 0.0  # e.g. NLB123866_01
    return unharmonicity

def computeAccented(song: Song, level, epsilon: float = 10e-4):
    return [True if song.mtcsong['features']['beatstrength'][ix] >= (level-epsilon) else False for ix in range(song.getSongLength())]


class ImpliedHarmony:

    def __init__(self, wpc: 'PitchContext'):
        self.wpc = wpc
        self.song = wpc.song
        self.songlength = self.song.getSongLength()
        self.params = wpc.params  
        self.numchords = 0 # will be set in self.initchords()
        self.chordquality = {} # will be set in self.initchords()
        self.initchords()

    def initchords(self):
        chordmask_dim = np.zeros(40, dtype=int)
        chordmask_min = np.zeros(40, dtype=int)
        chordmask_maj = np.zeros(40, dtype=int)
        chordmask_dom = np.zeros(40, dtype=int)

        chordmask_dim[[0, 11, 22]] = [1, 3, 5] #[root, third, fifth]
        chordmask_min[[0, 11, 23]] = [1, 3, 5] #[root, third, fifth]
        chordmask_maj[[0, 12, 23]] = [1, 3, 5] #[root, third, fifth]
        chordmask_dom[[0, 12, 23, 34]] = [1, 3, 5, 7] #[root, third, fifth, seventh]

        self.chordquality = {
            0: 'dim',
            1: 'm',
            2: '',
            3: '7'
        }
        self.masks = np.stack([chordmask_dim, chordmask_min, chordmask_maj, chordmask_dom])
        self.numchords = self.masks.shape[0]
        self.chordtones = np.zeros((40,40,self.numchords), dtype=int) #(pitch, root of chord, chord quality). 1: root, 3: third, 5: fifth, 7: seventh
        for chordq in range(self.numchords):
            for rootpitch in range(40):
                chordmask_shift = np.roll(self.masks, rootpitch, axis=1)
                self.chordtones[np.where(chordmask_shift[chordq]),rootpitch,chordq] = chordmask_shift[chordq][np.where(chordmask_shift[chordq])]

    def chordTransitionScore(
            self,
            chords,
            traceback,
            chord1_ixs,
            chord2_ixs,
            scalemask=np.ones(40, dtype=bool),
            song=None,
            wpc=None,
            ih=None
        ):
        # Basic implementation
        # Please, provide your own.

        #scoring scheme.
        pitch1 = chord1_ixs[1] % 40
        pitch2 = chord2_ixs[1] % 40
        #no score if root of chord tones is not in the scalemask)
        if not scalemask[pitch1] or not scalemask[pitch2]:
            return 0.0

        # start with score for 'next' chord
        score = chords[chord2_ixs]
        
        return score

    #ix = ix in full song
    def getP40(self, ix):
        return (self.song.mtcsong['features']['pitch40'][ix] - 1) % 40

    # if extendToAllNaturalTones, also include tones that are 'missing'. E.g. NLB070513_01 in G, but no F#
    # include all alterations. So, no F in melody, include Fb, F, and F#
    def getScaleMask(self, extendToAllNaturalTones=False):

        naturals = np.ones(7, dtype=bool) #[F C G D A E B] # True if stemtone not present in melody
        naturalixs = np.array([19, 2, 25, 8, 31, 14, 37])
        naturalsix = np.zeros(40, dtype=int)
        naturalsix[30:33] = 4 #A
        naturalsix[36:39] = 6 #B
        naturalsix[1:4]   = 1 #C
        naturalsix[7:10]  = 3 #D
        naturalsix[13:16] = 5 #E
        naturalsix[18:21] = 0 #F
        naturalsix[24:27] = 2 #G

        seq_length = len(self.wpc.ixs)

        #make scalemask for each note. includes the alteration of the stemtone that is CLOSEST to the focus note
        #TODO: different weight for looking back and looking forward?
        #TODO: different scalemasks for pre, post, and full contexts.

        #first record at which index the closest same stemtone is in the melody
        mostrecent = np.zeros((seq_length, 7), dtype=int) + 1000 #gives the index of the most recent stemtone (1000 if not yet)
        mostnext = np.zeros((seq_length, 7), dtype=int) - 1000 #gives the index of the closes next occurrence of the scale tone (-1000 if not any more)
        #first pitch for mostrecent
        p40 = self.getP40(self.wpc.ixs[0])
        mostrecent[0][naturalsix[p40]] = 0
        #rest
        for ix in range(1, seq_length):
            mostrecent[ix] = mostrecent[ix-1]
            p40 = self.getP40(self.wpc.ixs[ix])
            mostrecent[ix][naturalsix[p40]] = ix
        #last pitch for mostnext
        p40 = p40 = self.getP40(self.wpc.ixs[-1])
        mostnext[-1][naturalsix[p40]] = seq_length - 1
        #rest
        for ix in reversed(range(seq_length-1)):
            mostnext[ix] = mostnext[ix+1]
            p40 = p40 = self.getP40(self.wpc.ixs[ix])
            mostnext[ix][naturalsix[p40]] = ix

        #now create the scales for each of the notes
        scalemask = np.zeros((seq_length, 40))

        for ix in range(seq_length):
            diff_prev = np.abs(ix - mostrecent[ix])
            diff_next = np.abs(mostnext[ix] - ix)
            next_ixs = np.where(diff_prev > diff_next)[0]
            prev_ixs = np.where(diff_prev <= diff_next)[0]
            #from prev context
            for scale_ix in prev_ixs:
                if mostrecent[ix][scale_ix] != 1000:
                    p40 = self.getP40( self.wpc.ixs[ mostrecent[ix][scale_ix] ] )
                    scalemask[ix, p40] = True
                    naturals[naturalsix[p40]] = False
            #take from next context
            for scale_ix in next_ixs:
                if mostnext[ix][scale_ix] != -1000:
                    p40 = self.getP40( self.wpc.ixs[ mostnext[ix][scale_ix] ] )
                    scalemask[ix, p40] = True
                    naturals[naturalsix[p40]] = False

        # scalemask = np.zeros(40)
        # for ix in range(self.wpc.pitchcontext.shape[0]):
        #     p40 = (self.song.mtcsong['features']['pitch40'][ix] - 1) % 40
        #     scalemask[p40] = True
        #     naturals[naturalsix[p40]] = False

        # now extend each 
        # missing stem tones should be missing in all scalemasks, but addition might differ

        if extendToAllNaturalTones:

            for seq_ix in range(seq_length):

                #add missing tones
                #flats
                for ix in sorted(list(np.where(naturals)[0])): #from F->B order is important, more than one tone might be missing. first fix tones far away from C
                    #figure out whether to add flat

                    missing       = naturalixs[ix]
                    missingflat   = missing - 1 
                    fifthdown     = naturalixs[(ix - 1) % 7]
                    fifthdownflat = fifthdown - 1
                    fifthup       = naturalixs[(ix + 1) % 7]
                    fifthupflat   = fifthup - 1
                    fifthupsharp  = fifthup + 1

                    #### SOLVE MINOR, leading tone (if not in melody)

                    if ix == 0:
                        #When Fb? Only if Cb is present
                        if scalemask[seq_ix,fifthupflat]:
                            scalemask[seq_ix,missingflat] = True
                        #when F? If no C#
                        if not scalemask[seq_ix,fifthupsharp]:
                            scalemask[seq_ix,missing] = True
                        continue
                    
                    if ix == 6: #Bb ?
                        #When Bb? If Eb is present or if F is present
                        if scalemask[seq_ix,fifthdownflat] or scalemask[seq_ix,fifthup]:
                            scalemask[seq_ix,missingflat] = True
                        #when B? always - We deal with B# below
                        scalemask[seq_ix,missing] = True
                        continue

                    if scalemask[seq_ix,fifthdownflat]:
                        scalemask[seq_ix,missingflat] = True
                    elif scalemask[seq_ix,fifthdown] and scalemask[seq_ix,fifthup]:
                        scalemask[seq_ix,missing] = True
                    elif scalemask[seq_ix,fifthdown] and scalemask[seq_ix,fifthupflat]:
                        scalemask[seq_ix,missingflat] = True
                        scalemask[seq_ix,missing] = True

                #sharps
                for ix in sorted(list(np.where(naturals)[0]), reverse=True): #from B->F order is important, more than one tone might be missing. first fix tones far away from C
                    
                    #figure out whether to add flat

                    missing        = naturalixs[ix]
                    missingsharp   = missing + 1 
                    fifthdown      = naturalixs[(ix - 1) % 7]
                    fifthdownsharp = fifthdown + 1
                    fifthup        = naturalixs[(ix + 1) % 7]
                    fifthupsharp   = fifthup + 1
                    fifthupflat    = fifthup - 1

                    if ix == 6: # B#?
                        #when B#? If E# present
                        if scalemask[seq_ix,fifthdownsharp]:
                            scalemask[seq_ix,missingsharp] = True
                        #when B? when No flats (i.e. no Eb)
                        if not scalemask[seq_ix,fifthdownflat]:
                            scalemask[seq_ix,missing] = True
                        continue

                    if ix == 0: # F#?
                        #when F#? If C# present or if B present
                        if scalemask[seq_ix,fifthupsharp] or scalemask[seq_ix,fifthdown]:
                            scalemask[seq_ix,missingsharp] = True
                        #when F? If Bb not present (deal with Bb above)
                        if not scalemask[seq_ix,fifthdownflat]:
                            scalemask[seq_ix,missing] = True
                        continue

                    if scalemask[seq_ix,fifthupsharp]:
                        scalemask[seq_ix,missingsharp] = True
                    elif scalemask[seq_ix,fifthup] and scalemask[seq_ix,fifthdown]:
                        scalemask[seq_ix,missing] = True
                    elif scalemask[seq_ix,fifthdownsharp] and scalemask[seq_ix,fifthup]:
                        scalemask[seq_ix,missingsharp] = True
                        scalemask[seq_ix,missing] = True

        return scalemask

    def getGlobalScalemask(self, scalemask):
        return np.where(sum(scalemask) != 0, 1, 0)

    def getOptimalChordSequence(self, use_scalemask=True, chordTransitionScoreFunction=None):

        if chordTransitionScoreFunction == None:
            chordTransitionScoreFunction = self.chordTransitionScore

        if use_scalemask:
            scalemask = self.getScaleMask(extendToAllNaturalTones=True)
        else:
            scalemask = np.ones((len(self.wpc.ixs), 40), dtype=bool)
        
        chords      = self.getChords(use_scalemask=use_scalemask)
        numpitches  = chords.shape[1]        
        score       = np.zeros( (self.wpc.pitchcontext.shape[0], numpitches, self.numchords) )
        traceback   = np.zeros( (self.wpc.pitchcontext.shape[0], numpitches, self.numchords, 4), dtype=int ) #coordinates (pitch, chord, ix last root change, previous pitch) for previous chord
        trace       = np.zeros( (self.wpc.pitchcontext.shape[0], 2), dtype=int ) # (pitch, chord) for each note
        trace_score = np.zeros( (self.wpc.pitchcontext.shape[0], 2) ) # (score, score_diff) for each note

        #initialization: first note gets its own chords
        score[0] = chords[0]
        for ix in range(1, self.wpc.pitchcontext.shape[0]):
            #find indices of chord1
            chord1_ixs = np.where(chords[ix-1])
            #find indices of chord2
            chord2_ixs = np.where(chords[ix])

            #E.g., NLB191190_01, note 40?
            #if all available chords are forbidden (score -10.), another chord gets chosen (score 0)
            #TODO: all chord1_ixs should also be present in chord2_ixs, to always allow continuation of a chord?
            #TODO: better: only take maximum over scores in chord2_ixs
            
            for ixs2 in zip(chord2_ixs[0], chord2_ixs[1]):
                allscores = np.zeros( (numpitches, self.numchords) )
                for ixs1 in zip(chord1_ixs[0], chord1_ixs[1]):
                    transistionscore = chordTransitionScoreFunction(
                        chords,
                        traceback,
                        (ix-1,ixs1[0],ixs1[1]),
                        (ix, ixs2[0], ixs2[1]),
                        scalemask=scalemask[ix],
                        song=self.song,
                        wpc=self.wpc,
                        ih=self,
                    )
                    allscores[ixs1] = score[ix-1, ixs1[0], ixs1[1]] + transistionscore
                #now find max
                max_ixs = np.unravel_index(np.argmax(allscores), allscores.shape)
                maxscore = allscores[max_ixs]
                #update score
                score[ix, ixs2[0], ixs2[1]] = maxscore
                #update traceback
                traceback[ix, ixs2[0], ixs2[1]][0:2] = max_ixs
                if ixs2[0] % 40 != max_ixs[0] % 40: #new root @ ix
                    traceback[ix, ixs2[0], ixs2[1]][2] = ix
                    traceback[ix, ixs2[0], ixs2[1]][3] = max_ixs[0]
                else:
                    traceback[ix, ixs2[0], ixs2[1]][2] = traceback[ix-1, max_ixs[0], max_ixs[1]][2]  #take start of root from prvious in trace
                    traceback[ix, ixs2[0], ixs2[1]][3] = traceback[ix-1, max_ixs[0], max_ixs[1]][3]  #take previous pitch from prvious in trace


        #now do the traceback.
        #find max score for last note
        max_ixs = np.unravel_index(np.argmax(score[-1]), score[-1].shape)
        trace[-1] = (max_ixs[0], max_ixs[1])
        trace_score[-1] = (score[-1,max_ixs[0], max_ixs[1]], 0)
        for ix in range(self.wpc.pitchcontext.shape[0]-2, -1, -1):
            trace_ixs = traceback[ix+1,trace[ix+1][0],trace[ix+1][1]]
            thisscore = score[ix,trace_ixs[0], trace_ixs[1]]
            trace[ix] = (trace_ixs[0], trace_ixs[1])
            trace_score[ix][0] = thisscore

        for ix in range(1, self.wpc.pitchcontext.shape[0]):
            trace_score[ix][1] = trace_score[ix][0] - trace_score[ix-1][0]

        return trace, trace_score, score, traceback

    def trace2str(self, trace, contextType=False):
        if contextType:
            return [base40[trace[ix][0]%40] + self.chordquality[trace[ix][1]] + str(int(trace[ix][0]//40)) for ix in range(self.wpc.pitchcontext.shape[0])]
        else:
            return [base40[trace[ix][0]%40] + self.chordquality[trace[ix][1]]                              for ix in range(self.wpc.pitchcontext.shape[0])]

    def getChords(self, use_scalemask=True):
        seq_length = len(self.wpc.ixs)
        #prepare data structure:
        if use_scalemask:
            scalemask = self.getScaleMask(extendToAllNaturalTones=True)
        else:
            scalemask = np.ones((seq_length, 40), dtype=bool)
        numpitches = 120 # 40 pre, 40 post, 40 all
        chords = np.zeros( (self.wpc.pitchcontext.shape[0], numpitches, self.numchords ) ) # number of notes, 40 pre-pitches+40post-pitches, 4 chords (dim,min,maj,dom)
        for ix in range(self.wpc.pitchcontext.shape[0]):
            scores, strengths = self.getChordsForNote(self.wpc.pitchcontext[ix], normalize=True, scalemask=scalemask[ix])
            chords[ix] = np.multiply(
                scores,
                strengths
            )
        return chords

    def getChordsForNote(self, pitchcontextvector, normalize=True, scalemask=np.ones(40, dtype=bool)):
        #find out whether pitches could be arranged as series of thirds
        
        epsilon = 10e-4
        chordmask_minseventh = np.zeros(40)
        np.put(chordmask_minseventh, [34], 1.0) #used for check presence seventh in dom chord

        #only take natural tones, one b or one # as root
        valid_shifts = [1, 2, 3, 7, 8, 9, 13, 14, 15, 18, 19, 20, 24, 25, 26, 30, 31, 32, 36, 37, 38]

        #self.masks contains integers indicating the function of the tone in the chord (root, third, fifht, seventh)
        #here we need masks with ones everywhere
        binarymasks = np.clip(self.masks, 0, 1)

        #get a value for every rotation of the chordmasks
        score_pre = np.zeros((40, self.numchords))
        score_post = np.zeros((40, self.numchords))
        score_all = np.zeros((40, self.numchords))
        strength_pre = np.zeros((40, self.numchords))
        strength_post = np.zeros((40, self.numchords))
        strength_all = np.zeros((40, self.numchords))
        for shift in range(40):
            if not shift in valid_shifts:
                continue

            chordmask_shift = np.roll(binarymasks, shift, axis=1)
            chordmask_minseventh_shift = np.roll(chordmask_minseventh, shift)

            #only accept chords which have all notes in the (local) scale
            for maskid in range(self.numchords):
                if np.prod(scalemask[np.where(chordmask_shift[maskid])]) < epsilon:
                    chordmask_shift[maskid] = 0

            score_pre[shift] = np.sum(np.multiply(pitchcontextvector[:40],chordmask_shift), axis=1)
            if np.sum(pitchcontextvector[:40]) > epsilon:
                strength_pre[shift] = score_pre[shift] / np.sum(pitchcontextvector[:40])
            score_post[shift] = np.sum(np.multiply(pitchcontextvector[40:],chordmask_shift), axis=1)
            if np.sum(pitchcontextvector[40:]) > epsilon:
                strength_post[shift] = score_post[shift] / np.sum(pitchcontextvector[40:])
            score_all[shift] = np.sum(np.multiply(pitchcontextvector[:40]+pitchcontextvector[40:],chordmask_shift), axis=1)
            if np.sum(pitchcontextvector) > epsilon:
                strength_all[shift] = score_all[shift] / np.sum(pitchcontextvector)

            #if seventh in dom chord is not present -> erase dom chord
            if np.sum(np.multiply(chordmask_minseventh_shift,pitchcontextvector[:40])) < epsilon:
                score_pre[shift][3] = 0.0
                strength_pre[shift][3] = 0.0
            if np.sum(np.multiply(chordmask_minseventh_shift,pitchcontextvector[40:])) < epsilon:
                score_post[shift][3] = 0.0
                strength_post[shift][3] = 0.0
            if np.sum(np.multiply(chordmask_minseventh_shift,pitchcontextvector[:40]+pitchcontextvector[:40])) < epsilon:
                score_all[shift][3] = 0.0
                strength_all[shift][3] = 0.0

        if normalize:
            if np.sum(score_pre) > 0.0:
                score_pre  = score_pre / np.max(score_pre)
            if np.sum(score_post) > 0.0:
                score_post = score_post / np.max(score_post)
            if np.sum(score_all) > 0.0:
                score_all  = score_all / np.max(score_all)

        scores    = np.concatenate( (score_pre, score_post, score_all) )
        strengths = np.concatenate( (strength_pre, strength_post, strength_all) )

        return scores, strengths
    
    def printChordsForNote(self, chords):
        epsilon = 10e-4
        chordixs = np.where(chords > epsilon)
        chords_pre = []
        chords_post = []
        chords_all = []

        for ix in zip(chordixs[0], chordixs[1]):
            info = (base40[ix[0] % 40], self.chordquality[ix[1]], chords[ix])
            if ix[0] < 40:
                chords_pre.append(info)
            else:
                if ix[0] < 80:
                    chords_post.append(info)
                else:
                    chords_all.append(info)
        
        print('pre:')
        for info in sorted(chords_pre, key=lambda x: x[2], reverse=True):
            print(info)

        print('post:')
        for info in sorted(chords_post, key=lambda x: x[2], reverse=True):
            print(info)

        print('all:')
        for info in sorted(chords_all, key=lambda x: x[2], reverse=True):
            print(info)

    def printChord(self, chord_in):
        chordixs = np.where(chord_in > 0)
        chords = []
        
        for ix in zip(chordixs[0], chordixs[1]):
            pitch = ix[0] % 40
            chords.append((ix[0], base40[pitch], self.chordquality[ix[1]], chord_in[ (ix[0], ix[1]) ]))
        for chord in sorted(chords, key=lambda x: x[3], reverse=True):
            print(chord)
    
    #chords: output of getChords
    #transition from ix1 to ix2
    def printAllTransitions(self, chords, ix1, ix2, scores=None, chordTransitionScoreFunction=None, use_scalemask=True):
        if chordTransitionScoreFunction == None:
            chordTransitionScoreFunction = self.chordTransitionScore
        if scores is None:
            scores = np.zeros([chords.shape[0], chords.shape[1], self.numchords])
        
        #find indices of chord1
        chord1_ixs = np.where(chords[ix1])
        #find indices of chord2
        chord2_ixs = np.where(chords[ix2])

        if use_scalemask:
            scalemask = self.getScaleMask(extendToAllNaturalTones=True)
        else:
            scalemask = np.ones((len(self.wpc.ixs), 40), dtype=bool)

        transitions = []
        for ixs2 in zip(chord2_ixs[0], chord2_ixs[1]):
            for ixs1 in zip(chord1_ixs[0], chord1_ixs[1]):
                transistionscore = chordTransitionScoreFunction(
                    chords,
                    (ix1,ixs1[0],ixs1[1]),
                    (ix2, ixs2[0], ixs2[1]),
                    scalemask=scalemask[ix2],
                    song=self.song,
                    wpc=self.wpc
                )
                side1 = 'pre'
                if ixs1[0] > 39:
                    if ixs1[0] > 79:
                        side1 = 'all'
                    else:
                        side2 = 'post'
                side2 = 'pre'
                if ixs2[0] > 39:
                    if ixs2[0] > 79:
                        side2 = 'all'
                    else:
                        side2 = 'post'
                #report transition:
                transitions.append((
                    base40[ixs1[0] % 40],
                    self.chordquality[ixs1[1]],
                    side1,
                    '->',
                    base40[ixs2[0] % 40],
                    self.chordquality[ixs2[1]],
                    side2,
                    transistionscore,
                    scores[ix1, ixs1[0], ixs1[1]],
                    transistionscore + scores[ix1, ixs1[0], ixs1[1]],
                ))
        for tr in sorted(transitions, key=lambda x: x[9], reverse=True):
            print(' '.join([str(t) for t in tr]))

    def printTrace(self, trace, traceback):
        for ix, tr in enumerate(trace):
            print(
                ix,
                base40[trace[ix][0]%40],
                self.chordquality[trace[ix][1]],
                traceback[
                    ix,
                    trace[ix][0],
                    trace[ix][1]
                ]
            )




"""Provides methods for visualialization of data and results."""

import numpy as np
from matplotlib import pyplot as plt

#from datetime import datetime
#print(__file__, datetime.now().strftime("%H:%M:%S"))

#for repeated pitches:
#spans would be better for ixs [(start,end),(start,end),...]
def getColoredIxs(target_ixs, ixs, songlength):
    colorixs = []
    for ix in target_ixs:
        colorixs.append(ix)
        temp_ix = ix+1
        while (not temp_ix in ixs) and (temp_ix < songlength):
            colorixs.append(temp_ix)
            temp_ix += 1
    return colorixs

def array2colordict(array, ixs, criterion, songlength, color='red', greynan=True):
    ixsnp = np.array(ixs)
    target_ixs = ixsnp[np.where(criterion(array))]
    nan_ixs = []
    if greynan:
        nan_ixs = ixsnp[np.where(np.isnan(array))]        
    color_ixs = getColoredIxs(target_ixs, ixs, songlength)
    return {color:color_ixs, 'grey':nan_ixs}

#color all notes with high novelty
def novelty2colordict(novelty, ixs, percentile, songlength, color='red', greynan=True):
    criterion = lambda x : x >= np.nanpercentile(novelty,percentile)
    return array2colordict(
        novelty,
        ixs,
        criterion,
        songlength,
        color,
        greynan
    )

#color all notes with low consonance
def dissonance2colordict(dissonance, ixs, percentile, songlength, color='red', greynan=True):
    criterion = lambda x : x >= np.nanpercentile(dissonance,percentile)
    return array2colordict(
        dissonance,
        ixs,
        criterion,
        songlength,
        color,
        greynan
    )

#color all unharmonic notes
def unharmonicity2colordict(unharmonicity, ixs, threshold, songlength, color='red', greynan=True):
    criterion = lambda x : x >= threshold
    return array2colordict(
        unharmonicity,
        ixs,
        criterion,
        songlength,
        color,
        greynan
    )


def plotArray(array, ixs, xlabel : str, ylabel : str, figsize=(10,3)):
    fig, ax = plt.subplots(figsize=figsize)
    plt.ylim(-0.05,np.max(np.nan_to_num(array)) * 1.05)
    plt.plot(array)
    plt.xlabel(xlabel, fontsize='large')
    plt.ylabel(ylabel, fontsize='large')
    plt.xticks(np.arange(0, len(ixs), 1), [str(i) for i in ixs])
    plt.xticks(rotation = 90)
    return fig, ax